##############################################################################
#
# Copyright (c) 2010 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import tempfile
import logging
import os
import shutil
import unittest
import slapos.client
import six

try:
  import mock
except ImportError:
  from unittest import mock

from slapos.cli.prune import do_prune


class TestPrune(unittest.TestCase):
  def setUp(self):
    self.logger = mock.create_autospec(logging.getLogger())
    self.base_directory = tempfile.mkdtemp()
    self.shared_part_root = os.path.join(self.base_directory, 'shared')
    self.software_root = os.path.join(self.base_directory, 'software')
    self.instance_root = os.path.join(self.base_directory, 'instance')

    for d in (self.shared_part_root, self.software_root, self.instance_root):
      os.mkdir(d)
    self.addCleanup(shutil.rmtree, self.base_directory)

    self.config = {
        'shared_part_list': self.shared_part_root,
        'software_root': self.software_root,
        'instance_root': self.instance_root,
    }

  def _createFakeSoftware(self, name, using='', software_root=None):
    software_path = os.path.join(software_root or self.software_root, name)
    os.mkdir(software_path)
    with open(os.path.join(software_path, '.installed.cfg'), 'w') as f:
      f.write("""[buildout]
      using = {}
      """.format(using))
    return software_path

  def _createSharedPart(self, name, using='', shared_part_root=None):
    shared_part_name = os.path.join(shared_part_root or self.shared_part_root, name)
    shared_part_with_version = os.path.join(shared_part_name, name)
    os.makedirs(shared_part_with_version)
    with open(os.path.join(shared_part_with_version,
                           '.slapos.recipe.cmmi.signature'), 'w') as f:
      f.write("""signature
      using = {}
      """.format(using))
    return shared_part_with_version

  def test_simple_not_used_share_part(self):
    not_used = self._createSharedPart('not_used')
    used = self._createSharedPart('used_part')
    self._createFakeSoftware(self.id(), using=used)
    do_prune(self.logger, self.config, False)
    self.assertTrue(os.path.exists(used))
    self.assertFalse(os.path.exists(not_used))
    self.logger.warning.assert_called_with(
        'Unusued shared parts at %s%s', not_used, ' ... removed')

  def test_dry_run(self):
    not_used = self._createSharedPart('not_used')
    used = self._createSharedPart('used_part')
    self._createFakeSoftware(self.id(), using=used)
    do_prune(self.logger, self.config, True)
    self.assertTrue(os.path.exists(used))
    self.assertTrue(os.path.exists(not_used))
    self.logger.warning.assert_called_with(
        'Unusued shared parts at %s%s', not_used, '')

  def test_shared_part_used_in_another_shared_part(self):
    not_used = self._createSharedPart('not_used')
    indirectly_used_part = self._createSharedPart('part_used_indirectly')
    directly_used_part = self._createSharedPart('directly_used_part', using=indirectly_used_part)
    self._createFakeSoftware(self.id(), using=directly_used_part)
    do_prune(self.logger, self.config, False)
    self.assertTrue(os.path.exists(indirectly_used_part))
    self.assertTrue(os.path.exists(directly_used_part))
    self.assertFalse(os.path.exists(not_used))
    self.logger.warning.assert_called_with(
        'Unusued shared parts at %s%s', not_used, ' ... removed')

  def test_shared_part_not_used_recursive_dependencies(self):
    used_only_by_orphean_part = self._createSharedPart(
        'used_only_by_orphean_part')
    not_used = self._createSharedPart(
        'not_used', using=used_only_by_orphean_part)
    used_directly = self._createSharedPart('used_directly')

    self._createFakeSoftware(self.id(), using=used_directly)
    do_prune(self.logger, self.config, True)
    self.assertIn(
        mock.call('Unusued shared parts at %s%s', not_used, ''),
        self.logger.warning.mock_calls)
    self.assertIn(
        mock.call(
            'Unusued shared parts at %s%s', used_only_by_orphean_part, ''),
        self.logger.warning.mock_calls)
    do_prune(self.logger, self.config, False)
    self.assertTrue(os.path.exists(used_directly))
    self.assertFalse(os.path.exists(not_used))
    self.assertFalse(os.path.exists(used_only_by_orphean_part))

  def test_shared_part_used_in_buildout_script(self):
    not_used = self._createSharedPart('not_used')
    used_in_script = self._createSharedPart('used_in_script')
    fake_software_path = self._createFakeSoftware(self.id())
    os.mkdir(os.path.join(fake_software_path, 'bin'))
    script = os.path.join(fake_software_path, 'bin', 'buildout')
    with open(script, 'w') as f:
      f.write('#!{}'.format(used_in_script))
    # bin/ can also contain binary executables, this should not cause problems
    binary_script = os.path.join(fake_software_path, 'bin', 'binary')
    with open(binary_script, 'wb') as f:
      f.write(b'\x80')
    do_prune(self.logger, self.config, False)
    self.assertTrue(os.path.exists(used_in_script))
    self.assertFalse(os.path.exists(not_used))
    self.logger.warning.assert_called_with(
        'Unusued shared parts at %s%s', not_used, ' ... removed')
    if six.PY3:
      self.logger.debug.assert_any_call(
        'Skipping script %s that could not be decoded', binary_script)


  def test_shared_part_used_in_recursive_instance(self):
    used_in_software_from_instance = self._createSharedPart('used_in_software_from_instance')
    used_in_shared_part_from_instance = self._createSharedPart('used_in_shared_part_from_instance')
    not_used = self._createSharedPart('not_used')

    # create instance
    instance = os.path.join(self.instance_root, 'slappart0')
    instance_etc = os.path.join(instance, 'etc')
    instance_software = os.path.join(instance, 'software')
    instance_instance = os.path.join(instance, 'instance')
    instance_shared_part_root = os.path.join(instance, 'shared')
    for p in instance_etc, instance_software, instance_instance, instance_shared_part_root:
      os.makedirs(p)
    instance_slapos_cfg = os.path.join(instance_etc, 'slapos.cfg')
    with open(instance_slapos_cfg, 'w') as f:
      f.write('''
[slapos]
software_root = {instance_software}
instance_root = {instance_instance}
shared_part_list =
  {self.shared_part_root}
  {instance_shared_part_root}
'''.format(**locals()))

    # install software and shared part in instance
    software_in_instance = self._createFakeSoftware(
        'soft_in_instance',
        using=used_in_software_from_instance,
        software_root=instance_software
    )
    shared_part_in_instance = self._createSharedPart(
        'shared_part_in_instance',
        using=used_in_shared_part_from_instance,
        shared_part_root=instance_shared_part_root
    )
    unused_shared_part_in_instance = self._createSharedPart(
        'unused_shared_part_in_instance',
        shared_part_root=instance_shared_part_root
    ) # could be pruned, but prune is not recursive.

    do_prune(self.logger, self.config, False)
    self.logger.debug.assert_any_call(
        'Reading config at %s', instance_slapos_cfg)

    self.assertTrue(os.path.exists(used_in_software_from_instance))
    self.assertTrue(os.path.exists(used_in_software_from_instance))
    self.assertTrue(os.path.exists(software_in_instance))
    self.assertTrue(os.path.exists(shared_part_in_instance))
    self.assertFalse(os.path.exists(not_used))

    self.logger.warning.assert_called_with(
        'Unusued shared parts at %s%s', not_used, ' ... removed')

  def test_recursive_instance_broken_slapos_cfg(self):
    instance = os.path.join(self.instance_root, 'slappart0')
    instance_etc = os.path.join(instance, 'etc')
    os.makedirs(instance_etc)
    instance_slapos_cfg = os.path.join(instance_etc, 'slapos.cfg')

    # slapos.cfg might be not valid .ini file
    with open(instance_slapos_cfg, 'w') as f:
      f.write('this is not an actual slapos.cfg')
    do_prune(self.logger, self.config, False)
    self.logger.debug.assert_any_call(
        'Ignored config at %s because of error', instance_slapos_cfg, exc_info=True)

    # ... or an ini file without [slapos] section
    with open(instance_slapos_cfg, 'w') as f:
      f.write('[something]')
    do_prune(self.logger, self.config, False)
    self.logger.debug.assert_any_call(
        'Ignored config at %s because of error', instance_slapos_cfg, exc_info=True)

    # ... or referencing non existant paths
    with open(instance_slapos_cfg, 'w') as f:
      f.write('''
[slapos]
software_root = NOT EXIST SOFTWARE
instance_root = NOT EXIST INSTSTANCE
shared_part_list = NOT EXIST SHARED PART
''')
    do_prune(self.logger, self.config, False)
    self.assertIn(
        mock.call(
            'Ignoring non existant software root %s from %s',
            'NOT EXIST SOFTWARE',
            instance_slapos_cfg),
        self.logger.debug.mock_calls)
    self.assertIn(
        mock.call(
            'Ignoring non existant instance root %s from %s',
            'NOT EXIST INSTSTANCE',
            instance_slapos_cfg),
        self.logger.debug.mock_calls)

    self.assertIn(
        mock.call(
            'Ignoring non existant shared root %s from %s',
            'NOT EXIST SHARED PART',
            instance_slapos_cfg),
        self.logger.debug.mock_calls)


# -*- coding: utf-8 -*-
# vim: set et sts=2:
##############################################################################
#
# Copyright (c) 2018 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import os
import glob
import fnmatch
import sys
import logging
import time
import re
import json
import importlib
import traceback
import psutil
import inspect
import hashlib
from datetime import datetime
from multiprocessing import Process, Queue as MQueue
from six.moves import queue, reload_module
from slapos.util import str2bytes, mkdir_p, chownDirectory
from slapos.grid.utils import dropPrivileges, killProcessTree
from slapos.grid.promise import interface
from slapos.grid.promise.generic import (GenericPromise, PromiseQueueResult,
                                         AnomalyResult, TestResult,
                                         PROMISE_STATE_FOLDER_NAME,
                                         PROMISE_RESULT_FOLDER_NAME,
                                         PROMISE_HISTORY_RESULT_FOLDER_NAME,
                                         PROMISE_PARAMETER_NAME)
from slapos.grid.promise.wrapper import WrapPromise
from slapos.version import version

PROMISE_CACHE_FOLDER_NAME = '.slapgrid/promise/cache'

class PromiseError(Exception):
  pass

class PromiseProcess(Process):

  """
    Run a promise in a new Process
  """

  def __init__(self, partition_folder, promise_name, promise_path, argument_dict,
      logger, allow_bang=True, uid=None, gid=None, wrap=False, 
      check_anomaly=False):
    """
      Initialise Promise Runner

      @param promise_name: The name of the promise to run
      @param promise_path: path of the promise
      @param argument_dict: all promise parameters in a dictionary
      @param queue: Queue used to send promise result
      @param allow_bang: Bolean saying if bang should be called in case of
        anomaly failure.
      @param check_anomaly: Bolean saying if promise anomaly should be run.
      @param wrap: say if the promise should be wrapped in a subprocess using
        WrapPromise class
    """
    Process.__init__(self)
    # set deamon to True, so promise process will be terminated if parent exit
    self.daemon = True
    self.name = promise_name
    self.promise_path = promise_path
    self.logger = logger
    self.allow_bang = allow_bang
    self.check_anomaly = check_anomaly
    self.argument_dict = argument_dict
    self.uid = uid
    self.gid = gid
    self.partition_folder = partition_folder
    self.wrap_promise = wrap
    self._periodicity = None
    self.cache_folder = os.path.join(self.partition_folder,
                                     PROMISE_CACHE_FOLDER_NAME)
    self.cache_file = os.path.join(self.cache_folder, self.getPromiseTitle())
    # XXX - remove old files used to store promise timestamp and periodicity
    self._cleanupDeprecated()

  def _cleanupDeprecated(self):
    timestamp_file = os.path.join(self.partition_folder,
                                  PROMISE_STATE_FOLDER_NAME,
                                  '%s.timestamp' % self.name)
    periodicity_file = os.path.join(self.partition_folder,
                                    PROMISE_STATE_FOLDER_NAME,
                                    '%s.periodicity' % self.name)
    if os.path.exists(timestamp_file) and os.path.isfile(timestamp_file):
      os.unlink(timestamp_file)
    if os.path.exists(periodicity_file) and os.path.isfile(periodicity_file):
      os.unlink(periodicity_file)

  def getPromiseTitle(self):
    return os.path.splitext(self.name)[0]

  def updatePromiseCache(self, promise_class, promise_instance, started=True):
    """
      Cache some data from the promise that can be reused
    """
    py_file = '%s.py' % os.path.splitext(inspect.getfile(promise_class))[0]
    stat = os.stat(py_file)
    timestamp = time.time()
    cache_dict = dict(
      is_tested= not hasattr(promise_instance, 'isTested') or \
          promise_instance.isTested(),
      is_anomaly_detected=not hasattr(promise_instance, 'isAnomalyDetected') or \
          promise_instance.isAnomalyDetected(),
      periodicity=promise_instance.getPeriodicity(),
      next_run_after=timestamp + (promise_instance.getPeriodicity() * 60.0),
      timestamp=timestamp,
      module_file=py_file,
      module_file_mtime=stat.st_mtime,
    )
    if not started:
      cache_dict['next_run_after'] = timestamp
    with open(self.cache_file, 'w') as f:
      f.write(json.dumps(cache_dict))

  def loadPromiseCacheDict(self):
    """
      Load cached data for this promise.
      If saved promise module file is not exists then invalidate cache.
      Cache will be updated when promise run
    """
    if os.path.exists(self.cache_file):
      try:
        with open(self.cache_file) as f:
          cache_dict = json.loads(f.read())
          if not os.path.exists(cache_dict['module_file']):
            # file not exists mean path was changed
            return None
          return cache_dict
      except ValueError:
        return None

  def run(self):
    """
      Run the promise
      
      This will first load the promise module (which will update process sys.path)
    """
    try:
      os.chdir(self.partition_folder)
      promise_started = False
      if self.uid and self.gid:
        dropPrivileges(self.uid, self.gid, logger=self.logger)
      mkdir_p(self.cache_folder)
      if self.wrap_promise:
        promise_instance = WrapPromise(self.argument_dict)
      else:
        self._createInitFile()
        promise_module = self._loadPromiseModule()
        promise_instance = promise_module.RunPromise(self.argument_dict)

      if not hasattr(promise_instance, 'isAnomalyDetected') or not \
          hasattr(promise_instance, 'isTested') or \
          (promise_instance.isAnomalyDetected() and self.check_anomaly) or \
          (promise_instance.isTested() and not self.check_anomaly):
        # if the promise will run, we save execution timestamp
        promise_started = True
      self.updatePromiseCache(
        WrapPromise if self.wrap_promise else promise_module.RunPromise,
        promise_instance,
        started=promise_started)
      promise_instance.run(self.check_anomaly, self.allow_bang)
    except Exception:
      self.logger.error(traceback.format_exc())
      raise

  def _createInitFile(self):
    promise_folder = os.path.dirname(self.promise_path)
    # if there is no __init__ file, add it
    init_file = os.path.join(promise_folder, '__init__.py')
    if not os.path.exists(init_file):
      with open(init_file, 'w') as f:
        f.write("")
      os.chmod(init_file, 0o644)
    # add promise folder to sys.path so we can import promise script
    if sys.path[0] != promise_folder:
      sys.path[0:0] = [promise_folder]

  def _loadPromiseModule(self):
    """Load a promise from promises directory."""

    if re.match(r'[a-zA-Z_]', self.name) is None:
      raise ValueError("Promise plugin name %r is not valid" % self.name)

    promise_module = importlib.import_module(os.path.splitext(self.name)[0])
    if not hasattr(promise_module, "RunPromise"):
      raise AttributeError("Class RunPromise not found in promise" \
        "%s" % self.name)
    if not interface.IPromise.implementedBy(promise_module.RunPromise):
      raise RuntimeError("RunPromise class in %s must implement 'IPromise'"
        " interface. @implementer(interface.IPromise) is missing ?" % self.name)

    from slapos.grid.promise.generic import GenericPromise
    if not issubclass(promise_module.RunPromise, GenericPromise):
      raise RuntimeError("RunPromise class is not a subclass of " \
        "GenericPromise class.")

    if 'py'.join(promise_module.__file__.rsplit('pyc')) != self.promise_path:
      # cached module need to be updated
      promise_module = reload_module(promise_module)
    # load extra parameters
    self._loadPromiseParameterDict(promise_module)

    return promise_module

  def _loadPromiseParameterDict(self, promise_module):
    """Load a promise parameters."""
    if hasattr(promise_module, PROMISE_PARAMETER_NAME):
      extra_dict = getattr(promise_module, PROMISE_PARAMETER_NAME)
      if not isinstance(extra_dict, dict):
        raise ValueError("Extra parameter is not a dict")
      for key in extra_dict:
        if key in self.argument_dict:
          raise ValueError("Extra parameter name %r cannot be used.\n%s" % (
                           key, extra_dict))
        self.argument_dict[key] = extra_dict[key]


class PromiseLauncher(object):

  def __init__(self, config=None, logger=None, dry_run=False):
    """
      Promise launcher will run promises

      @param config_file: A file containing configurations
      @param dry_run: Only run all promises without save the result
      @param logger: Set the logger to use, if None a logger will be configured
        to console.
      @param config: A configuration dict to use. Values send here will
        overwrite configs from `config_file`. Expected values in config are:
        promise-timeout
          Maximum promise execution time before timeout. Default: 20
        partition-folder
          Base path of the partition
        promise-folder
          Promises folder, all promises scripts will be imported from that folder
        legacy-promise-folder
          Legacy promises folder, where to find bash, shell and standard promises
        log-folder
          Folder where promises will write logs. Can be None
        check-anomaly
          Ask to check anomaly instead of test. Default: False
        debug
          Configure loggin in debug mode. Default: True
        master-url
          SlapOS Master service URL
        partition-cert
          Computer Partition Certificate file
        partition-key
          Computer Partition key file
        partition-id
          Computer Partition ID, ex: slappart13
        computer-id
          Computer ID, ex: COMP-1234
        uid
          User UID
        gid
          User GID
        debug
          If True, show Promise consumption and execution time information, etc
        run-only-promise-list
          A list of promise from plugins directory that will be executed
        force
          Set to True if force run promises without check their periodicity
    """
    self.dry_run = dry_run
    self.__config = {
      'promise-timeout': 20,
      'promise-folder': None,
      'legacy-promise-folder': None,
      'log-folder': None,
      'partition-folder': None,
      'debug': False,
      'uid': None,
      'gid': None,
      'master-url': None,
      'partition-cert': None,
      'partition-key': None,
      'partition-id': None,
      'computer-id': None,
      'check-anomaly': False,
      'force': False,
      'run-only-promise-list': None
    }
    if config is not None:
      self.__config.update(config)

    for key, value in self.__config.items():
      setattr(self, key.replace('-', '_'), value or None)

    if self.promise_folder is None:
      raise ValueError("Promise folder is missing in configuration!")
    if self.partition_folder is None:
      raise ValueError("Partition folder is missing in configuration!")

    if logger is None:
      self.logger = logging.getLogger(__name__)
      self.logger.setLevel(logging.DEBUG if self.debug else logging.INFO)
      if len(self.logger.handlers) == 0 or \
          not isinstance(self.logger.handlers[0], logging.StreamHandler):
        handler = logging.StreamHandler()
        handler.setFormatter(
          logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        )
        self.logger.addHandler(handler)
    else:
      self.logger = logger

    self.queue_result = MQueue()
    self.bang_called = False
    self._skipped_amount = 0

    self.promise_output_dir = os.path.join(
      self.partition_folder,
      PROMISE_RESULT_FOLDER_NAME
    )
    if not os.path.exists(self.promise_output_dir):
      mkdir_p(self.promise_output_dir)
      self._updateFolderOwner()

    self.promise_history_output_dir = os.path.join(
      self.partition_folder,
      PROMISE_HISTORY_RESULT_FOLDER_NAME
    )
    if not os.path.exists(self.promise_history_output_dir):
      mkdir_p(self.promise_history_output_dir)
      self._updateFolderOwner()

  def _generatePromiseResult(self, promise_process, promise_name, promise_path,
      message, execution_time=0):
    if self.check_anomaly:
      problem = False
      promise_result = self._loadPromiseResult(promise_process.getPromiseTitle())
      if promise_result is not None and (promise_result.item.hasFailed() or
          'error:' in promise_result.item.message.lower()):
        # generate failure if latest promise result was error
        # If a promise timeout it will return failure if the timeout occur again
        problem = True
      result = AnomalyResult(problem=problem, message=message)
    else:
      result = TestResult(problem=True, message=message)
    return PromiseQueueResult(
      item=result,
      path=promise_path,
      name=promise_name,
      title=promise_process.getPromiseTitle(),
      execution_time=execution_time
    )

  def _savePromiseResult(self, result):
    if not isinstance(result, PromiseQueueResult):
      self.logger.error('Bad result: %s is not type of PromiseQueueResult...' % result)
      return

    promise_output_file = os.path.join(
      self.promise_output_dir,
      "%s.status.json" % result.title
    )
    promise_tmp_file = '%s.tmp' % promise_output_file

    with open(promise_tmp_file, "w") as outputfile:
      json.dump(result.serialize(), outputfile)
    os.rename(promise_tmp_file, promise_output_file)

  def _saveGlobalState(self, state_dict, config, report_date, success, error):
    promise_status_file = os.path.join(self.partition_folder,
                                       PROMISE_STATE_FOLDER_NAME,
                                       'promise_status.json')
    global_file = os.path.join(self.partition_folder,
                               PROMISE_STATE_FOLDER_NAME,
                               'global.json')
    public_file = os.path.join(self.partition_folder,
                               PROMISE_STATE_FOLDER_NAME,
                               'public.json')
    global_state_dict = dict(
      status="ERROR" if error else "OK",
      state={"error": error, "success": success},
      type='global',
      portal_type='Software Instance',
      title='Instance Monitoring',
      date=report_date,
      _links={}, # To be filled by the instance
      data={'state': 'monitor_state.data',
            'process_state': 'monitor_process_resource.status',
            'process_resource': 'monitor_resource_process.data',
            'memory_resource': 'monitor_resource_memory.data',
            'io_resource': 'monitor_resource_io.data',
            'monitor_process_state': 'monitor_resource.status'},
      aggregate_reference=config.get('computer-id'),
      partition_id=config.get('partition-id')
      )

    public_state_dict = dict(
      status=global_state_dict["status"],
      date=report_date,
      _links={}
    )
    with open(promise_status_file, "w") as f:
      json.dump(state_dict, f)

    with open(global_file, "w") as f:
      json.dump(global_state_dict, f)

    with open(public_file, "w") as f:
      json.dump(public_state_dict, f)

  def _savePromiseHistoryResult(self, result):
    state_dict = result.serialize()
    previous_state_dict = {}
    promise_status_file = os.path.join(self.partition_folder,
                                       PROMISE_STATE_FOLDER_NAME,
                                       'promise_status.json')

    if os.path.exists(promise_status_file):
      with open(promise_status_file) as f:
        try:
          previous_state_dict = json.load(f)
        except ValueError:
          pass

    history_file = os.path.join(
      self.promise_history_output_dir,
      '%s.history.json' % result.title
    )

    # Remove useless informations
    result_dict = state_dict.pop('result')
    result_dict["change-date"] = result_dict["date"]
    state_dict.update(result_dict)
    state_dict.pop('path', '')
    state_dict.pop('type', '')
    state_dict["status"] = "ERROR" if result.item.hasFailed() else "OK"

    if not os.path.exists(history_file) or not os.stat(history_file).st_size:
      with open(history_file, 'w') as f:
        f.write("""{
          "date": %s,
          "data": %s}""" % (time.time(), json.dumps([state_dict])))
    else:
      previous_state_list = previous_state_dict.get(result.name, None)
      if previous_state_list is not None:
        _, change_date, checksum = previous_state_list
        current_sum = hashlib.md5(str2bytes(state_dict.get('message', ''))).hexdigest()
        if state_dict['change-date'] == change_date and \
            current_sum == checksum:
          # Only save the changes and not the same info
          return
  
      state_dict.pop('title', '')
      state_dict.pop('name', '')
      with open (history_file, mode="r+") as f:
        f.seek(0,2)
        f.seek(f.tell() -2)
        f.write('%s}' % ',{}]'.format(json.dumps(state_dict)))

  def _saveStatisticsData(self, stat_file_path, date, success, error):
    # csv-like document for success/error statictics
    if not os.path.exists(stat_file_path) or os.stat(stat_file_path).st_size == 0:
      with open(stat_file_path, 'w') as fstat:
        fstat.write("""{
          "date": %s,
          "data": ["Date, Success, Error, Warning"]}""" % time.time())
  
    current_state = '%s, %s, %s, %s' % (
        date,
        success,
        error,
        '')
  
    # append to file
    # XXX this is bad, it is getting everywhere.
    if current_state:
      with open (stat_file_path, mode="r+") as fstat:
        fstat.seek(0,2)
        position = fstat.tell() -2
        fstat.seek(position)
        fstat.write('%s}' % ',"{}"]'.format(current_state))

  def _loadPromiseResult(self, promise_title):
    promise_output_file = os.path.join(
      self.promise_output_dir,
      "%s.status.json" % promise_title
    )
    result = None
    if os.path.exists(promise_output_file):
      with open(promise_output_file) as f:
        try:
          result = PromiseQueueResult()
          result.load(json.loads(f.read()))
        except ValueError as e:
          result = None
          self.logger.warn('Bad promise JSON result at %r: %s' % (
            promise_output_file,
            e
          ))
    return result

  def _writePromiseResult(self, result_item):
    if result_item.item.type() == "Empty Result":
      # no result collected (sense skipped)
      return
    elif result_item.item.hasFailed():
      self.logger.error(result_item.item.message)
      if result_item.execution_time != -1 and \
          isinstance(result_item.item, AnomalyResult) and self.check_anomaly:
        # stop to bang as it was called
        self.bang_called = True
    # Send result
    self._savePromiseResult(result_item)
    self._savePromiseHistoryResult(result_item)

  def _emptyQueue(self):
    """Remove all entries from queue until it's empty"""
    while True:
      try:
        self.queue_result.get_nowait()
      except queue.Empty:
        return

  def _updateFolderOwner(self, folder_path=None):
    stat_info = os.stat(self.partition_folder)
    if folder_path is None:
      folder_path = os.path.join(self.partition_folder,
                                 PROMISE_STATE_FOLDER_NAME)
    chownDirectory(folder_path, stat_info.st_uid, stat_info.st_gid)

  def isPeriodicityMatch(self, next_timestamp):
    if next_timestamp:
      return time.time() >= next_timestamp
    return True

  def _launchPromise(self, promise_name, promise_path, argument_dict,
      wrap_process=False):
    """
      Launch the promise and save the result. If promise_module is None,
      the promise will be run with the promise process wap module.

      If the promise periodicity doesn't match, the previous promise result is
      checked.

      Returns a TestResult or None if promise never ran.
    """
    try:
      promise_process = PromiseProcess(
        self.partition_folder,
        promise_name,
        promise_path,
        argument_dict,
        logger=self.logger,
        check_anomaly=self.check_anomaly,
        allow_bang=not (self.bang_called or self.dry_run) and self.check_anomaly,
        uid=self.uid,
        gid=self.gid,
        wrap=wrap_process,
      )

      promise_cache_dict = promise_process.loadPromiseCacheDict()
      if promise_cache_dict is not None:
        if self.check_anomaly and not promise_cache_dict.get('is_anomaly_detected') \
            or not self.check_anomaly and not promise_cache_dict.get('is_tested'):
          # promise is skipped, send empty result
          self._writePromiseResult(PromiseQueueResult())
          self._skipped_amount += 1
          return
      if not self.force and (promise_cache_dict is not None and not
          self.isPeriodicityMatch(promise_cache_dict.get('next_run_after'))):
        # we won't start the promise process, just get the latest result
        self._skipped_amount += 1
        result = self._loadPromiseResult(promise_process.getPromiseTitle())
        if result is not None:
          if result.item.hasFailed():
            self.logger.error(result.item.message)
            return result.item
        return
      # we can do this because we run processes one by one
      # we cleanup queue in case previous result was written by a killed process
      self._emptyQueue()
      promise_process.start()
    except Exception:
      # only print traceback to not prevent run other promises
      self.logger.error(traceback.format_exc())
      self.logger.warning("Promise %s skipped." % promise_name)
      return TestResult(problem=True, message=traceback.format_exc())

    self.logger.info("Checking promise %s..." % promise_name)
    queue_item = None
    sleep_time = 0.05
    increment_limit = int(self.promise_timeout / sleep_time)
    execution_time = self.promise_timeout
    ps_profile = False
    if self.debug:
      try:
        psutil_process = psutil.Process(promise_process.pid)
        ps_profile = True
      except psutil.NoSuchProcess:
        # process is gone
        pass
    for current_increment in range(0, increment_limit):
      if not promise_process.is_alive():
        try:
          queue_item = self.queue_result.get(True, 1)
        except queue.Empty:
          # no result found in process result Queue
          pass
        else:
          queue_item.execution_time = execution_time
        break

      if ps_profile:
        try:
          io_counter = psutil_process.io_counters()
          self.logger.debug(
            "[t=%ss] CPU: %s%%, MEM: %s MB (%s%%), DISK: %s Read - %s Write" % (
              current_increment*sleep_time,
              psutil_process.cpu_percent(),
              psutil_process.memory_info().rss / float(2 ** 20),
              round(psutil_process.memory_percent(), 4),
              io_counter.read_count,
              io_counter.write_count
            )
          )
        except (psutil.AccessDenied, psutil.NoSuchProcess):
          # defunct process will raise AccessDenied
          pass
      time.sleep(sleep_time)
      execution_time = (current_increment + 1) * sleep_time
    else:
      promise_process.terminate()
      promise_process.join(1) # wait for process to terminate
      # if the process is still alive after 1 seconds, we kill it
      if promise_process.is_alive():
        self.logger.info("Killing process %s..." % promise_name)
        killProcessTree(promise_process.pid, self.logger)

      message = 'Error: Promise timed out after %s seconds' % self.promise_timeout
      queue_item = self._generatePromiseResult(
        promise_process,
        promise_name=promise_name,
        promise_path=promise_path,
        message=message,
        execution_time=execution_time
      )

    if queue_item is None:
      queue_item = self._generatePromiseResult(
        promise_process,
        promise_name=promise_name,
        promise_path=promise_path,
        message="Error: No output returned by the promise",
        execution_time=execution_time
      )

    self._writePromiseResult(queue_item)
    if self.debug:
      self.logger.debug("Finished promise %r in %s second(s)." % (
                       promise_name, execution_time))

    return queue_item.item

  def run(self):
    """
      Run all promises
    """
    promise_list = []
    failed_promise_name = ""
    failed_promise_output = ""
    previous_state_dict = {}
    new_state_dict = {}
    report_date = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S+0000')
    promise_status_file = os.path.join(self.partition_folder,
                     PROMISE_STATE_FOLDER_NAME,
                     'promise_status.json')

    promise_result_file = os.path.join(self.partition_folder,
                     PROMISE_STATE_FOLDER_NAME,
                     'promise_result.json')

    promise_stats_file = os.path.join(self.partition_folder,
                     PROMISE_STATE_FOLDER_NAME,
                     'promise_stats.json')

    if os.path.exists(promise_status_file):
      with open(promise_status_file) as f:
        try:
          previous_state_dict = json.load(f)
        except ValueError:
          pass

    new_state_dict = previous_state_dict.copy()
    base_config = {
      'log-folder': self.log_folder,
      'partition-folder': self.partition_folder,
      'debug': self.debug,
      'promise-timeout': self.promise_timeout,
      'master-url': self.master_url,
      'partition-cert': self.partition_cert,
      'partition-key': self.partition_key,
      'partition-id': self.partition_id,
      'computer-id': self.computer_id,
      'queue': self.queue_result,
      'slapgrid-version': version,
    }
    error = 0
    success = 0
    promise_name_list = []
    if os.path.exists(self.promise_folder) and os.path.isdir(self.promise_folder):
      for promise_name in os.listdir(self.promise_folder):
        for suffix in ['.pyc', '.pyo']:
          if promise_name.endswith(suffix):
            promise_path = os.path.join(self.promise_folder, promise_name)
            if not os.path.exists(promise_path[:-1]):
              try:
                os.unlink(promise_path)
              except Exception as e:
                self.logger.warning('Failed to remove %r because of %s', promise_path, e)
              else:
                self.logger.debug('Removed stale %r', promise_path)

        if promise_name.startswith('__init__') or \
            not promise_name.endswith('.py'):
          continue

        promise_name_list.append(promise_name)
        if self.run_only_promise_list is not None and not \
            promise_name in self.run_only_promise_list:
          continue

        promise_path = os.path.join(self.promise_folder, promise_name)
        config = {
          'path': promise_path,
          'name': promise_name
        }
        config.update(base_config)
        promise_result = self._launchPromise(promise_name, promise_path, config)
        if promise_result:
          change_date = promise_result.date.strftime('%Y-%m-%dT%H:%M:%S+0000')
          if promise_result.hasFailed():
            promise_status = 'FAILED'
            error += 1
          else:
            promise_status = "OK"
            success += 1
          if promise_name in previous_state_dict:
            status, previous_change_date, _ = previous_state_dict[promise_name]
            if promise_status == status:
              change_date = previous_change_date

          message = promise_result.message if promise_result.message else ""
          new_state_dict[promise_name] = [
            promise_status,
            change_date,
            hashlib.md5(str2bytes(message)).hexdigest()]
      
          if promise_result.hasFailed() and not failed_promise_name:
            failed_promise_name = promise_name
            failed_promise_output = promise_result.message
        else:
          # The promise was skip, so for statistic point of view we preserve
          # the previous result
          if promise_name in new_state_dict:
            if new_state_dict[promise_name][0] == "FAILED":
              error += 1
            else:
              success += 1

    if not self.run_only_promise_list and len(promise_name_list) > 0:
      # cleanup stale json files
      promise_output_dir_content = glob.glob(os.path.join(self.promise_output_dir, '*.status.json'))
      promise_history_output_dir_content = glob.glob(os.path.join(self.promise_history_output_dir, '*.history*.json'))
      for promise_file_name in promise_name_list:
        promise_name = promise_file_name[:-3]
        promise_status_json_name = promise_name + '.status.json'
        promise_history_json_match = promise_name + '.history*.json'
        promise_output_dir_content = [q for q in promise_output_dir_content if os.path.basename(q) != promise_status_json_name]
        promise_history_output_dir_content = [q for q in promise_history_output_dir_content if not fnmatch.fnmatch(os.path.basename(q), promise_history_json_match)]

      promise_output_dir_cleanup = [os.path.join(self.promise_output_dir, q) for q in promise_output_dir_content]
      promise_history_output_dir_cleanup = [os.path.join(self.promise_history_output_dir, q) for q in promise_history_output_dir_content]
      for path in promise_output_dir_cleanup + promise_history_output_dir_cleanup:
        try:
          os.unlink(path)
        except Exception:
          self.logger.exception('Problem while removing stale file %s', path)
        else:
          self.logger.info('Removed stale file %s', path)
      # drop old promises from new_state_dict
      for key in list(new_state_dict.keys()):
        if key not in promise_name_list:
          new_state_dict.pop(key, None)

    if not self.run_only_promise_list and os.path.exists(self.legacy_promise_folder) \
        and os.path.isdir(self.legacy_promise_folder):
      # run legacy promise styles
      for promise_name in os.listdir(self.legacy_promise_folder):
        promise_path = os.path.join(self.legacy_promise_folder, promise_name)
        if not os.path.isfile(promise_path) or \
            not os.access(promise_path, os.X_OK):
          self.logger.warning("Bad promise file at %r." % promise_path)
          continue

        config = {
          'path': promise_path,
          'name': promise_name
        }
        config.update(base_config)
        # We will use promise wrapper to run this
        promise_result = self._launchPromise(promise_name,
                                           promise_path,
                                           config,
                                           wrap_process=True)
        if promise_result:
          change_date = promise_result.date.strftime('%Y-%m-%dT%H:%M:%S+0000')
          if promise_result.hasFailed():
            promise_status = 'FAILED'
            error += 1
          else:
            promise_status = "OK"
            success += 1
          if promise_name in previous_state_dict:
            status, previous_change_date, _ = previous_state_dict[promise_name]
            if promise_status == status:
              change_date = previous_change_date

          message = promise_result.message if promise_result.message else ""
          new_state_dict[promise_name] = [
            promise_status,
            change_date,
            hashlib.md5(str2bytes(message)).hexdigest()]

          if promise_result.hasFailed() and not failed_promise_name:
            failed_promise_name = promise_name
            failed_promise_output = promise_result.message
        else:
          # The promise was skip, so for statistic point of view we preserve
          # the previous result
          if promise_name in new_state_dict:
            if new_state_dict[promise_name][0] == "FAILED":
              error += 1
            else:
              success += 1

    self._updateFolderOwner(self.promise_output_dir)

    self._saveStatisticsData(promise_stats_file,
                         report_date, success, error)

    self._saveGlobalState(new_state_dict, base_config, report_date, 
                          success, error) 

    if self._skipped_amount > 0:
      self.logger.info("%s promises didn't need to be checked." % \
        self._skipped_amount)
    if failed_promise_name:
      raise PromiseError("Promise %r failed with output: %s" % (
          failed_promise_name,
          failed_promise_output))




# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2010-2014 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from slapos.cli.config import ClientConfigCommand
from slapos.client import init, ClientConfig


class RemoveCommand(ClientConfigCommand):
    """
    remove a Software from a node
    """

    def get_parser(self, prog_name):
        ap = super(RemoveCommand, self).get_parser(prog_name)

        ap.add_argument('software_url',
                        help='Your software url')

        ap.add_argument('node',
                        help="Target node")

        return ap

    def take_action(self, args):
        configp = self.fetch_config(args)
        conf = ClientConfig(args, configp)
        local = init(conf, self.app.log)
        do_remove(self.app.log, args.software_url, args.node, local)


def do_remove(logger, software_url, computer_id, local):
    """
    Request deletion of Software Release
    'software_url' from computer 'computer_id'.
    """
    logger.info('Requesting deletion of %s Software Release...', software_url)

    if software_url in local:
        software_url = local[software_url]
    local['slap'].registerSupply().supply(
        software_release=software_url,
        computer_guid=computer_id,
        state='destroyed'
    )
    logger.info('Done.')



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2010-2014 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import logging

from slapos.cli.config import ConfigCommand
from slapos.proxy import do_proxy, ProxyConfig


class ProxyStartCommand(ConfigCommand):
    """
    minimalist, stand-alone SlapOS Master
    """

    def get_parser(self, prog_name):
        ap = super(ProxyStartCommand, self).get_parser(prog_name)

        ap.add_argument('-u', '--database-uri',
                        help='URI for sqlite database')
        ap.add_argument('--port',
                        help='Port to use')
        ap.add_argument('--host',
                        help='Host to use')

        return ap

    def take_action(self, args):
        configp = self.fetch_config(args)

        conf = ProxyConfig(logger=self.app.log)

        conf.mergeConfig(args, configp)

        if not self.app.options.log_file and hasattr(conf, 'log_file'):
            # no log file is provided by argparser,
            # we set up the one from config
            file_handler = logging.FileHandler(conf.log_file)
            formatter = logging.Formatter(self.app.LOG_FILE_MESSAGE_FORMAT)
            file_handler.setFormatter(formatter)
            self.app.log.addHandler(file_handler)

        conf.setConfig()

        do_proxy(conf=conf)



##############################################################################
#
# Copyright (c) 2013 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import os
import pkg_resources
import pyflakes.scripts.pyflakes
import sys
import unittest

class CheckCodeConsistency(unittest.TestCase):
  """Lints all SlapOS Node and SLAP library code base."""
  def setUp(self):
    self._original_argv = sys.argv
    sys.argv = [sys.argv[0],
                os.path.join(
                    pkg_resources.get_distribution('slapos.core').location,
                    'slapos',
                )
               ]

  def tearDown(self):
    sys.argv = self._original_argv

  @unittest.skip('pyflakes test is disabled')
  def testCodeConsistency(self):
    if pyflakes.scripts.pyflakes.main.func_code.co_argcount:
      pyflakes.scripts.pyflakes.main([
                os.path.join(
                    pkg_resources.get_distribution('slapos.core').location,
                    'slapos',
                )])
    else:
      pyflakes.scripts.pyflakes.main()




# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import sys
import glob
import os
import six.moves.configparser as configparser

from slapos.cli.command import check_root_user
from slapos.cli.config import ConfigCommand
from slapos.grid.slapgrid import merged_options
from slapos.grid.utils import setRunning, setFinished
from slapos.util import rmtree, string_to_boolean


class PruneCommand(ConfigCommand):
  """Clean up unused shared slapos.recipe.cmmi parts.

  This simple script does not detect inter-dependencies and needs to run
  multiple times. For example, if A depend on B, when B is not used, B will
  be removed on first run and A will then become unused.
  """
  command_group = 'node'

  def get_parser(self, prog_name):
    ap = super(PruneCommand, self).get_parser(prog_name)
    ap.add_argument(
        '--dry-run', help="Don't delete, just log", action='store_true')
    return ap

  def take_action(self, args):
    configp = self.fetch_config(args)
    options = merged_options(args, configp)

    if not options.get('shared_part_list'):
      self.app.log.error('No shared_part_list options in slapos config')
      sys.exit(-1)

    if string_to_boolean(options.get('root_check', 'True').lower()):
      check_root_user(self)

    pidfile_software = options.get('pidfile_software')
    if not args.dry_run and pidfile_software and os.path.exists(
        pidfile_software):
      self.app.log.error('Cannot prune while software is running')
      sys.exit(-1)

    if pidfile_software:
      setRunning(logger=self.app.log, pidfile=pidfile_software)
    try:
      do_prune(self.app.log, options, args.dry_run)
    finally:
      if pidfile_software:
        setFinished(pidfile_software)


def _prune(
    logger,
    shared_root,
    software_root,
    instance_root,
    ignored_shared_parts,
    dry_run,
):
  signatures = getUsageSignatureFromSoftwareAndSharedPart(
      logger, software_root, shared_root, ignored_shared_parts)

  # recursively look in instance
  signatures.update(getUsageSignaturesFromSubInstance(logger, instance_root))

  for shared_part in glob.glob(os.path.join(shared_root, '*', '*')):
    if shared_part not in ignored_shared_parts:
      logger.debug("checking shared part %s", shared_part)
      h = os.path.basename(shared_part)
      for soft, installed_cfg in signatures.items():
        if h in installed_cfg:
          logger.debug("It is used in %s", soft)
          break
      else:
        if not dry_run:
          rmtree(shared_part)
        logger.warning(
            'Unusued shared parts at %s%s', shared_part,
            '' if dry_run else ' ... removed')
        yield shared_part


def _prune_loop(logger, shared_root, software_root, instance_root, dry_run):
  ignored_shared_parts = set([])
  while True:
    pruned = list(
        _prune(
            logger,
            shared_root,
            software_root,
            instance_root,
            ignored_shared_parts,
            dry_run,
        ))
    ignored_shared_parts.update(pruned)
    if not pruned:
      break


def do_prune(logger, options, dry_run):
  shared_root = options['shared_part_list'].splitlines()[-1].strip()
  logger.warning("Pruning shared directories at %s", shared_root)
  _prune_loop(
      logger,
      shared_root,
      options['software_root'],
      options['instance_root'],
      dry_run,
  )


def getUsageSignaturesFromSubInstance(logger, instance_root):
  """Look at instances in instance_root to find used shared parts,
  if instances are recursive slapos.

  The heuristic is that if an instance contain a file named slapos.cfg,
  this is a recursive slapos.
  """
  signatures = {}
  for slapos_cfg in getInstanceSlaposCfgList(logger, instance_root):
    cfg = readSlaposCfg(logger, slapos_cfg)
    if not cfg:
      return {}
    shared_root = None
    if cfg['shared_part_list']:
      shared_root = cfg['shared_part_list'][-1]
      if not os.path.exists(shared_root):
        logger.debug(
            "Ignoring non existant shared root %s from %s",
            shared_root,
            slapos_cfg,
        )
        shared_root = None

    if not os.path.exists(cfg['software_root']):
      logger.debug(
          "Ignoring non existant software root %s from %s",
          cfg['software_root'],
          slapos_cfg,
      )
    else:
      signatures.update(
          getUsageSignatureFromSoftwareAndSharedPart(
              logger, cfg['software_root'], shared_root))
    if not os.path.exists(cfg['instance_root']):
      logger.debug(
          "Ignoring non existant instance root %s from %s",
          cfg['instance_root'],
          slapos_cfg,
      )
    else:
      signatures.update(
          getUsageSignaturesFromSubInstance(logger, cfg['instance_root']))
  return signatures


def getInstanceSlaposCfgList(logger, instance_root):
  """Find all slapos.cfg from instance directory, as instance
  can contain recursive slapos (that refer parts from outer slapos).
  """
  for root, _, filenames in os.walk(instance_root):
    if 'slapos.cfg' in filenames:
      yield os.path.join(root, 'slapos.cfg')


def readSlaposCfg(logger, path):
  """Read a slapos.cfg found in an instance directory.
  """
  logger.debug('Reading config at %s', path)
  parser = configparser.ConfigParser({'shared_part_list': ''})
  try:
    parser.read([path])
    cfg = {
      'software_root': parser.get('slapos', 'software_root'),
      'instance_root': parser.get('slapos', 'instance_root'),
      'shared_part_list': parser.get('slapos', 'shared_part_list').splitlines()
    }
  except configparser.Error:
    logger.debug('Ignored config at %s because of error', path, exc_info=True)
    return None
  logger.debug('Read config: %s', cfg)
  return cfg


def getUsageSignatureFromSoftwareAndSharedPart(
    logger, software_root, shared_root, ignored_shared_parts=None):
  """Look in all softwares and shared parts to collect the signatures
  that are used.
  `ignored_shared_parts` is useful during dry-run, we want to ignore
  already the parts that we are about to delete.
  """
  if ignored_shared_parts is None:
    ignored_shared_parts = set([])
  signatures = {}
  for installed_cfg in glob.glob(os.path.join(software_root, '*',
                                              '.installed.cfg')):
    with open(installed_cfg) as f:
      signatures[installed_cfg] = f.read()
  for script in glob.glob(os.path.join(software_root, '*', 'bin', '*')):
    with open(script) as f:
      try:
        signatures[script] = f.read()
      except UnicodeDecodeError:
        logger.debug("Skipping script %s that could not be decoded", script)
  if shared_root:
    for shared_signature in glob.glob(os.path.join(shared_root, '*', '*',
                                                   '.*signature')):
      if not any(shared_signature.startswith(ignored_shared_part)
                 for ignored_shared_part in ignored_shared_parts):
        with open(shared_signature) as f:
          signatures[shared_signature] = f.read()
  return signatures
